/***************************************************************************
 *   Copyright (C) 2004 by Fred Schaettgen                                 *
 *   kde.sch@ttgen.net                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
#include "clockpaintview.h"
#include <kmessagebox.h>
#include <algorithm>
#include <qtimer.h>
#include <qdir.h>
#include <qdatetime.h>
#include <kjs/interpreter.h>
#include <kdebug.h>
#include "ecmafunc.h"
#include <qfile.h>
#include <qtextstream.h>
#include <klocale.h>
#include <kapplication.h>
#include <kstandarddirs.h>
#include "prefs.h"
#include <qtooltip.h>

using namespace std;

const QString resourceDirName = "styleclock";

ClockPaintView::ClockPaintView(QWidget* parent, Prefs* prefs) :
    QGLWidget(parent),
    envSeconds(0),
    envMinutes(0),
    envHours(0),
    prefs(prefs)
{   
    interpreter = NULL;
        
    defineLayerFunc = new ECMAFunc(this);
    connect(defineLayerFunc, SIGNAL(call(KJS::ExecState*, KJS::Object&, 
        const KJS::List&, KJS::Value&)), this, SLOT(ecmaSlotDefineLayer( 
            KJS::ExecState*, KJS::Object&, const KJS::List&, KJS::Value&)));
    
    timer = new QTimer(this);
    connect(timer, SIGNAL(timeout()), this, SLOT(slotTimeout()));
    
    globalObject = KJS::Object(new KJS::ObjectImp());
    interpreter = new KJS::Interpreter(globalObject);
    reloadSettings();
}

ClockPaintView::~ClockPaintView()
{
    delete interpreter;
}

int ClockPaintView::loadTexture(QString filename)
{
    QImage buf;
    buf.setAlphaBuffer(true);
    if ( ! buf.load(filename) ) {  
        KMessageBox::sorry(this, "Failed to load image", "Failed");
        return 0;
    }
        
    GLuint texName;
    glGenTextures(1, &texName);
    glBindTexture(GL_TEXTURE_2D, texName);
    
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
    
    int wOld, hOld;
    int mipmapLevel = 0;
    do {
        QImage tex1;
        wOld = buf.width();
        hOld = buf.height();
        tex1 = QGLWidget::convertToGLFormat(buf);
        
        glTexImage2D(GL_TEXTURE_2D, mipmapLevel, 4, tex1.width(), tex1.height(), 0,
            GL_RGBA, GL_UNSIGNED_BYTE, tex1.bits() );
        buf = buf.smoothScale(max(1, buf.width()/2), max(1, buf.height()/2));
        ++mipmapLevel;
    
    } while (wOld != buf.width() && hOld != buf.height());
    return texName;
} 

void ClockPaintView::initializeGL()
{
    // Set up the rendering context, define display lists etc.:
    glClearColor( 0.0, 0.0, 0.0, 0.0 );
    glDisable(GL_DEPTH_TEST);
    glEnable(GL_BLEND);
    glDisable(GL_DITHER);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_TEXTURE_2D);

    glGenTextures(1, &bgTexName);
    //loadTexture("/usr/share/icons/crystalsvg/128x128/apps/clock.png");
}

void ClockPaintView::resizeGL( int w, int h )
{
    // setup viewport, projection etc.:
    this->screenWidth = w;
    this->screenHeight = h;
    glViewport( 0, 0, (GLint)w, (GLint)h );
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    float left = 0;
    float right = w;
    float top = 0;
    float bottom = h;
    float near = 0;
    float far = 1;
    glOrtho(left, right, bottom, top, near, far);
}

void ClockPaintView::paintGL()
{
    // Get the background pixmap
        
    QPixmap bgPixmap(screenWidth, screenHeight);
    bgPixmap.fill(QColor(255,0,0));
    bgPixmap.fill(parentWidget(), 0, 0);
    double bgOrigSizeX = bgPixmap.width();
    double bgOrigSizeY = bgPixmap.height();
    int bgExpandedSizeX = 1;
    int bgExpandedSizeY = 1;
    while (bgExpandedSizeX < bgOrigSizeX) bgExpandedSizeX *= 2; 
    while (bgExpandedSizeY < bgOrigSizeY) bgExpandedSizeY *= 2; 
    double bgTexRight = bgOrigSizeX / bgExpandedSizeX;
    double bgTexBottom = bgOrigSizeY / bgExpandedSizeY;
    bgPixmap.resize(bgExpandedSizeX, bgExpandedSizeY);
    //bgPixmap.fill(QColor(128,129,130));
    QImage bgImage = bgPixmap.convertToImage();
    bgImage.setAlphaBuffer(true);
    QImage bgGlImage = QGLWidget::convertToGLFormat(bgImage);
        
    glBindTexture(GL_TEXTURE_2D, bgTexName);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
    glTexImage2D(GL_TEXTURE_2D, 0, 4, bgGlImage.width(), bgGlImage.height(), 0,
            GL_RGBA, GL_UNSIGNED_BYTE, bgGlImage.bits() );
            
    // draw the scene:
    glMatrixMode(GL_MODELVIEW);
    //QColor bgColor = KApplication::palette().active().background();
    //glColor3d(bgColor.red()/256.0, bgColor.green()/256.0, bgColor.blue()/256.0);
    glBindTexture(GL_TEXTURE_2D, bgTexName);
    glEnable(GL_TEXTURE_2D);
    glBegin( GL_QUADS );
    glTexCoord2f(bgTexRight, 1.0);
    glVertex3f(screenWidth, 0., 0);
    glTexCoord2f(0, 1.0);
    glVertex3f(0. , 0., 0);
    glTexCoord2f(0, 1.0-bgTexBottom);
    glVertex3f(0., screenHeight,0);
    glTexCoord2f(bgTexRight, 1.0-bgTexBottom);
    glVertex3f(screenWidth, screenHeight, 0);
    glEnd();
    // Get rid of the background texture
        
    glColor4f(1, 1, 1, 1);
    
    if (interpreter == NULL) return;
    
    list<DisplayListItem>::const_iterator it;
    KJS::ExecState* state = interpreter->globalExec();
    
    for (it = displayList.begin(); it != displayList.end(); ++it) {
        
        DisplayListItem item = *it;
        double moveX = item.properties.get(state, "moveX").toNumber(state);
        double moveY = item.properties.get(state, "moveY").toNumber(state);
        double anchorX = item.properties.get(state, "anchorX").toNumber(state);
        double anchorY = item.properties.get(state, "anchorY").toNumber(state);
        double rotate = item.properties.get(state, "rotate").toNumber(state);
        double scaleX = item.properties.get(state, "scaleX").toNumber(state);
        double scaleY = item.properties.get(state, "scaleY").toNumber(state);
        /*kdDebug() << QString("Move(%1/%2) ").arg(moveX).arg(moveY) <<
            QString("Move(%1/%2) ").arg(moveX).arg(moveY) <<
            QString("Anchor(%1/%2) ").arg(anchorX).arg(anchorY) << 
            QString("Rotate(%1) ").arg(rotate) <<
            QString("Scale(%1/%2) ").arg(scaleX).arg(scaleY)<<  endl;*/
        
        //kdDebug() << "Drawing item " << item.imageName << endl;
        int textureName;
        if (imageMap.find(item.imageName) != imageMap.end()) {
            textureName = imageMap[item.imageName];
        }
        else {
            textureName = loadTexture(expandFilename(item.imageName));
            imageMap[item.imageName] = textureName;
        }
        //kdDebug() << item.imageName << " " << textureName << endl;
        glPushMatrix();
        
        glTranslatef(moveX, moveY, 0);
        glScalef(scaleX, scaleY, 1);
        
        glTranslatef(anchorX, anchorY, 0);
        glRotatef(rotate, 0, 0, 1);
        glTranslatef(-anchorX, -anchorY, 0);
        
        
        glBindTexture(GL_TEXTURE_2D, textureName);
        glEnable(GL_TEXTURE_2D);
        glBegin( GL_QUADS );
        glTexCoord2f(0, 1);
        glVertex3f(0, 0, 0);
        glTexCoord2f(1, 1);
        glVertex3f(1, 0, 0);
        glTexCoord2f(1, 0);
        glVertex3f(1, 1, 0);
        glTexCoord2f(0, 0);
        glVertex3f(0, 1,0);
        glEnd();
        glPopMatrix();
    }
}

int ClockPaintView::heightForWidth(int width)
{
    kdDebug() << "Calling heightForWidth(" << width << ")" << endl;
    if (interpreter == NULL) return width;
    
    executeScript();
    
    KJS::ExecState* state = interpreter->globalExec();
    KJS::Object globalObject = interpreter->globalObject();
    if (globalObject.hasProperty(state, "heightForWidth")) {
        kdDebug() << "Found heightForWidth property" << endl;
        KJS::Object h4wObj = globalObject.get(state, "heightForWidth").toObject(state);
        if (h4wObj.isNull() == false && h4wObj.implementsCall()) {
            KJS::List argList;
            argList.append(KJS::Number(width));
            KJS::Object thisObj(new KJS::ObjectImp());
            KJS::Value h = h4wObj.call(state, thisObj, argList);
            kdDebug() << "Height for width is " << h.toInteger(state) << endl;
            return max(8,min(256,h.toInteger(state)));   
        }
    }
    return width;
}

int ClockPaintView::widthForHeight(int height)
{
    kdDebug() << "Calling widthForHeight(" << height << ")" << endl;
    if (interpreter == NULL) return height;
    
    executeScript();
    
    KJS::ExecState* state = interpreter->globalExec();
    KJS::Object globalObject = interpreter->globalObject();
    if (globalObject.hasProperty(state, "widthForHeight")) {
        kdDebug() << "Found widthForHeight property" << endl;
        KJS::Object w4hObj = globalObject.get(state, "widthForHeight").toObject(state);
        if (w4hObj.isNull() == false && w4hObj.implementsCall()) {
            KJS::List argList;
            argList.append(KJS::Number(height));
            KJS::Object thisObj(new KJS::ObjectImp());
            KJS::Value w = w4hObj.call(state, thisObj, argList);
            kdDebug() << "Width for height is " << w.toInteger(state) << endl;
            return max(8,min(256,w.toInteger(state)));   
        }
    }
    return height;
}

void ClockPaintView::executeScript()
{
    if (interpreter == NULL) return;
    
    KJS::ExecState* state = interpreter->globalExec();
    
    // Updating the script state variables
    QDateTime curTime = QDateTime::currentDateTime();
    globalObject.put(state, "widthForHeight", KJS::Null());
    globalObject.put(state, "heightForWidth", KJS::Null());
    globalObject.put(state, KJS::Identifier("second"), KJS::Number(curTime.time().second()));
    globalObject.put(state, KJS::Identifier("minute"), KJS::Number(curTime.time().minute()));
    globalObject.put(state, KJS::Identifier("hour"), KJS::Number(curTime.time().hour()));
    globalObject.put(state, KJS::Identifier("screenHeight"),KJS::Number(screenHeight));
    globalObject.put(state, KJS::Identifier("screenWidth"),KJS::Number(screenWidth));
    
    globalObject.put(state, KJS::Identifier("showSeconds"), KJS::Boolean(showSeconds));
    
    displayList.clear();
    KJS::Completion c = interpreter->evaluate((const char*)(script.utf8()));
    QString msg;
    switch (c.complType()) {
    case KJS::Normal:
        msg = "Normal";
        break;
    case KJS::Break:
        msg = "Break";
        break;
    case KJS::Continue:
        msg = "continue";
        break;
    case KJS::ReturnValue:
        msg = "ReturnValue";
        break;
    case KJS::Throw:
        msg = QString("Throw: %1").arg(c.value().toString(state).ascii());
        break;
    default:
        msg = "Unknown result code";
    }
    //KMessageBox::information(this, msg, "Script result");
    if (msg!="Normal") kdDebug() << msg << endl;

}

void ClockPaintView::slotTimeout()
{
    executeScript();    
        
    /*KJS::Object global = interpreter->globalObject();
    if (global.hasProperty(state, KJS::Identifier("paint"))) {
        kdDebug() << "property paint found" << endl;
        KJS::Object paintObj = global.get(state, KJS::Identifier("paint")).toObject(state);
        KJS::List params;
        KJS::Value ret = paintObj.call(state, scriptThis, params);
        kdDebug() << ret.toString(state).ascii() << endl;
    }
    else {
        kdDebug() << "property paint NOT found" << endl;    
    }*/
    
    repaint();

    QDate curDate = QDate::currentDate();
    QString dateStr = curDate.toString();
    if (lastDateString != dateStr) {
        QToolTip::remove(this);
        QToolTip::add(this, dateStr);
        lastDateString = dateStr;
    } 
}

QString ClockPaintView::expandFilename(const QString& s)
{
    KStandardDirs *dirs = KApplication::kApplication()->dirs();
    return dirs->findResource("data", QString("%1/themes/%2/%3")
        .arg(resourceDirName).arg(themeName).arg(s));
}

QStringList ClockPaintView::getAvailableThemes()
{
    KStandardDirs *dirs = KApplication::kApplication()->dirs();
    QStringList themeList;
    QStringList themeDirs;
    themeDirs = dirs->findDirs("data", QString("%1/themes").arg(resourceDirName));
    for (int nDir=0; nDir < themeDirs.size(); ++nDir) {
        kdDebug() << "Searching for themes in " << themeDirs[nDir] << endl;
        QDir curDir(themeDirs[nDir]);
        QStringList themes = curDir.entryList(QDir::Dirs | QDir::Readable, QDir::Name);
        for (int nTheme=0; nTheme < themes.size(); ++nTheme) {
            kdDebug() << "Found theme " << themes[nTheme] << endl;
            if (themes[nTheme].startsWith(".")) continue;
            if (themeList.contains(themes[nTheme]) == 0) {
                themeList.push_back(themes[nTheme]);
            }
        }
    }
    return themeList;
}


void ClockPaintView::setTheme(const QString& s)
{
    timer->stop();
    themeName = s;
    displayList.clear();
    
    // Remove the old cached images
    std::map<QString, int>::iterator it;
    for (it = imageMap.begin(); it != imageMap.end(); ++it) {
        GLuint texture = it->second;
        glDeleteTextures(1, &texture);
    }
    imageMap.clear();
    
    //globalObject = KJS::Object(new KJS::ObjectImp());
    //delete interpreter;
    //interpreter = new KJS::Interpreter(globalObject);
    KJS::ExecState* state = interpreter->globalExec();
    
    globalObject.put(state, KJS::Identifier("defineLayer"), 
        KJS::Object(defineLayerFunc->getObjectImp()));
    
    QString scriptName = expandFilename("main.js");
    if (scriptName.length() > 0) {
        QFile f(scriptName);
        f.open(IO_ReadOnly);
        QTextStream stream(&f);
        script = stream.read();
    }
    else {
        KMessageBox::error(this, i18n("Failed to load script file %2 for theme %1")
            .arg(themeName).arg(scriptName));
        script = "";
        return;
    }

    timer->start(1000);
}

void ClockPaintView::ecmaSlotDefineLayer(KJS::ExecState */*exec*/, KJS::Object &/*obj*/, 
    const KJS::List &args, KJS::Value &ret)
{
    KJS::ExecState* state = interpreter->globalExec();
    int numArgs = args.size();
   
    // Default values for the parameters
    //item.moveX = item.moveY = item.anchorX = item.anchorY = item.rotate = 0;
    //item.scaleX = item.scaleY = 1;
    
    if (numArgs == 0 || numArgs >= 2) {
        kdError() << "Wrong number of arguments for js function paint." << endl;
        ret = KJS::Undefined();
        return;
    }
    
    QString imageName = QString(args[0].toString(state).ascii());
    DisplayListItem item(imageName);
    ret = item.properties;
    displayList.push_back(item);
    
    // Setting default values for the ecmascript object
    KJS::Object& props = item.properties;
    props.put(state, "moveX", KJS::Number(0));
    props.put(state, "moveY", KJS::Number(0));
    props.put(state, "anchorX", KJS::Number(0));
    props.put(state, "anchorY", KJS::Number(0));
    props.put(state, "rotate", KJS::Number(0));
    props.put(state, "scaleX", KJS::Number(1));
    props.put(state, "scaleY", KJS::Number(1));
    
    //kdDebug() << QString("called ecmaSlotDefineLayer(\"%1\")").arg(imageName) << endl;
}

ClockPaintView::DisplayListItem::DisplayListItem(QString imageName)
{
    this->imageName = imageName;
    properties = KJS::Object(new KJS::ObjectImp());
}

void ClockPaintView::reloadSettings()
{
    kdDebug() << "ClockPaintView::reloadSettings" << endl;
    
    if (prefs->theme() != themeName) {
        kdDebug() << "ClockPaintView: setting theme to " << prefs->theme() << endl;
        setTheme(prefs->theme());
    }
    showSeconds = prefs->showSeconds();
    emit sizeSuggestionChanged();
}


ClockPaintView::DisplayListItem::~DisplayListItem()
{
    //delete properties.imp();
}




#include "clockpaintview.moc"
